/* This is a "base" driver for USB.
 * It is not as comprehensive as my other "base" drivers, becase
 * USB devices can be a lot different from each other.
 * If you want in-depth examples, download some of my other USB drivers
 * from
 *  http://www.bolthole.com/solaris/drivers/
 *
 * What this driver does is attach to any USB ID you specify,
 * and print out endpoint information.
 *
 *    Philip Brown,  November 2001
 *
 *  %A%
 */

#include <varargs.h>
#define _SYS_VARARGS_H		/* sys/varargs.h hack for gcc */

#include <sys/usb/usba.h>

#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/stream.h>
#include <stropts.h>

/* THESE MUST BE LAST of the system defines*/
#include <sys/ddi.h>
#include <sys/sunddi.h>

#include "ov7xx0.h" /* camera register info */


/* I'm not exactly sure what a 'safe' depth to set is. 
 * We could fill up kernel memory very fast.
 * So keep it at a nice "safe" value, like 10
 */
#define AIPTEK_QUEUE_DEPTH 2

#define AIPTEK_MAX_TRANSFER 1000 /* XXX SET THIS SMALLER LATER!! */

/* As well as changing the NAME of this, you will also want to change
 * the TYPE of the minor node created.
 */
#define DEVNODE_NAME	"aipcam"	
#define DEVNODE_TYPE	DDI_NT_SERIAL /* see sunddi.h */

/**********************************************************************/
/*             Hardware defines and notes                             */
/**********************************************************************/

 /* these are not ACTUAL endpoint numbers, but convenience things,
  * in case I ever decide to add support for the OV511plus hardware
  */
#define AIPTEK_ENDPOINT_993		1
#define AIPTEK_ENDPOINT_961		2 /* 511+ only */
#define AIPTEK_ENDPOINT_769		3
#define AIPTEK_ENDPOINT_513		4
#define AIPTEK_ENDPOINT_257		5
#define AIPTEK_ENDPOINT_129		6 /* 511+ only */
#define AIPTEK_ENDPOINT_33		7 /* 511+ only */

  /***************************************************************/
  /* Now start the actual device registers that we need to tweak */
  /***************************************************************/
#define OV511_REG_PXCNT			0x12 /* "clamped pixel num" */
#define OV511_REG_LNCNT			0x13 /* "clamped line num" */
#define OV511_REG_PXDIV			0x14 /* "pixel divisor" */
#define OV511_REG_LNDIV			0x15 /* "line divisor" */
#define OV511_REG_UV_EN			0x16 /* color width/type selector */

#define OV511_REG_LSTR			0x17
#define OV511_REG_OPTS			0x18 /* vidcam 'options' */
 #define OV511_OPTS_YFILT		0x01 /* Ychannel low pass filter */

#define OV511_REG_SNAP_PXCNT		0x1a /* "clamped pixel num" */
#define OV511_REG_SNAP_LNCNT		0x1b /* "clamped line num" */
#define OV511_REG_SNAP_PXDIV		0x1c /* "pixel divisor" for stills*/
#define OV511_REG_SNAP_LNDIV		0x1d /* "line divisor"  for stills */
#define OV511_REG_SNAP_UV_EN		0x16 /* color width/type selector */
#define OV511_REG_SNAP_OPTS		0x1f /* snapshot 'options' */
 #define OV511_SNAPOPTS_YFILT		0x01 /* Ychannel low pass filter */


#define OV511_REG_ENFC			0x20 /* ENable Flow Control */
 #define OV511_ENFC_ON			0x01

#define OV511_REG_FIFOSIZE		0x30 /* match endpoint_xxx size */
#define OV511_REG_FIFO_OPTS		0x31
 #define OV511_FIFOOPT_NOZ		0x01 /* no compression */
 #define OV511_FIFOOPT_PKN		0x02 /* add extra byte for packetnum */
 #define OV511_FIFOOPT_ZPKG		0x08 /* add null EOF packet */

#define OV511_REG_I2C			0x40 /* control reg */
/* first, values for READING I2C control reg */
 #define  I2C_TMOUT			(1<<2) /* op has timed out */
 #define  I2C_NACK			(1<<1) /* check AFTER idle set */
 #define  I2C_IDLE			(1)    /* bus idle */
/* Now, bits you WRITE to the register */
 #define  I2C_ABORT			(1<<4)
 #define  I2C_3B_WT			(0<<1) /* 3byte write */
 #define  I2C_2B_WT			(1<<1) /* 2byte write */
 #define  I2C_2B_RD			(2<<1) /* 2byte read */
 /* write type is bits 0x6 */
 #define  I2C_START			(1)

#define OV511_REG_I2C_SID		0x41 /* 'slave ID' for writes */
					     /* What is this for?? */

#define OV511_REG_I2C_SWA		0x42 /* 3byte write 'address' */
#define OV511_REG_I2C_SMA		0x43 /* 2byte write 'address' */
#define OV511_REG_I2C_SRA		0x44 /* slave ID for read addr */
#define OV511_REG_I2C_DATA		0x45 /* Data port */
#define OV511_REG_I2C_CLOCK		0x46 /* I2C bus "prescalar" */


#define OV511_REG_RESET			0x50
/* values for RESET register, above */
 #define OV511_RESET_UDC		0x01
 #define OV511_RESET_I2C		0x02
 #define OV511_RESET_ISO_FIFO		0x04
 #define OV511_RESET_OMNI_CE		0x08
 #define OV511_RESET_DRAM		0x10
 #define OV511_RESET_CAMERA		0x20
 #define OV511_RESET_OV511		0x40

 #define OV511_RESET_STOPCAM		0x3d
 #define OV511_RESET_MOST		0x3f
 #define OV511_RESET_ALL		0x7f

#define OV511_REG_CLKDIV		0x51
#define OV511_REG_SNAP			0x52 /* hardware 'snapshot' register */
#define OV511_REG_ENSYS			0x53 /* ENable SYStem */

#define OV511_REG_LED			0x55
#define OV511_REG_USERREG		0x5e /* user-defined register */
#define OV511_REG_CUST_ID		0x5f

#define OV511_REG_CMP_EN		0x78 /* Compression Enable */
#define OV511_REG_LUT_EN		0x79 /* color Look-up table enable */
 #define OV511_LUT_Y_ENABLE		0x01 /* Y-channel */
 #define OV511_LUT_UV_ENABLE		0x02 /* UV-channel */


/************************************************************/
/* List of "alternate endpoints" for OV511                  */
/* alt1 = packetsize of 993                                 */
/* alt2 = packetsize of 768                                 */
/* alt3 = packetsize of 769                                 */
/* alt4 = packetsize of 512                                 */
/* alt5 = packetsize of 513                                 */
/* alt6 = packetsize of 257                                 */
/* alt7 = packetsize of 0(?)                                */


/**********************************************************************/
/**********************************************************************/

/* This ASSUMES you create minor nodes where the higher bits denote
 * the instance of the driver. (aka "which card is this?")
 * The lower bits denote clone number, or anything else you want.
 */

#define DEVT_TO_INST(devt) (getminor(devt)>>4)
#define DEVT_TO_CLONENUM(devt) (getminor(devt) & 0xf)

/* Used to choose a node  num for a device node compatible with DEVT_TO_INST*/
#define INST_TO_NODENUM(instance) (instance<<4)


/*Our globally visible stuffs, that we populate the global
 * driver entry structs with  [base_cb_ops, base_ops]
 * Yeah, its kind of wierd to have a "static" be also externally visible,
 * but that's traditionally how these are declared.
 */
static	int	aipcam_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
		    void **result);
static	int	aipcam_probe(dev_info_t *dip);
static	int	aipcam_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static	int	aipcam_open(dev_t *dev, int openflags, int otyp, cred_t *credp);

static	int	aipcam_print(dev_t, char*);

static	int	aipcam_read(dev_t dev, struct uio* ustr, cred_t * credp);

static	int	aipcam_close(dev_t dev, int openflags, int otyp, cred_t *credp);
static	int	aipcam_write(dev_t dev, struct uio *uiop, cred_t *credp);
static	int	aipcam_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
			   cred_t *credp, int *rvalp);
static	int	aipcam_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);


static int debug_level;

/* "char/block operations" */
static struct cb_ops	aipcam_cb_ops = {
	aipcam_open,		/* open() */
	aipcam_close,	/* close() */
	nodev,			/* strategy()		*/
	aipcam_print,	/* print routine	*/
	nodev,			/* no dump routine	*/
	aipcam_read,		/* read() */
	aipcam_write,	/* write() */
	aipcam_ioctl,	/* generic ioctl */
	nodev,			/* no devmap routine	*/
	nodev,			/* no mmap routine	*/
	nodev,			/* no segmap routine	*/
	nochpoll,		/* no chpoll routine	*/
	ddi_prop_op,
	0,			/* we are not a STREAMS driver	*/
	D_NEW | D_MP|D_64BIT,	/* safe for multi-thread/multi-processor */
	CB_REV,			/* cb_ops version? */
	nodev,			/* cb_aread() */
	nodev,			/* cb_awrite() */
};

/* "device operations". */
static struct dev_ops aipcam_ops = {
	DEVO_REV,		/* devo_rev */
	0,			/* devo_refcnt */
	aipcam_getinfo,		/* devo_getinfo */
	nulldev,		/* devo_identify */
	aipcam_probe,		/* devo_probe */
	aipcam_attach,		/* devo_attach */
	aipcam_detach,		/* devo_detach */
	nodev,			/* devo_reset */
	&aipcam_cb_ops,		/* devo_cb_ops */
	(struct bus_ops *)0,	/* devo_bus_ops */
	NULL,			/* devo_power */
};

/* Ignore this. This is only a filler for the next struct, and
 * is already declared in <sys/modctl.h>
 */
extern	struct	mod_ops mod_driverops;

static	struct modldrv modldrv = {
	&mod_driverops,
	"USB test driver",
	&aipcam_ops,
};

static	struct modlinkage modlinkage = {
	MODREV_1,		/* MODREV_1 is indicated by manual */
	{&modldrv,NULL,NULL,NULL}

};


/************************************************************
 * And here's finally the vars specific to THIS driver,
 * that get allocated on a per-instance basis.
 * Here's where you should keep instance-specific mutexes,
 * DMA handles, register mapping handles, and stuff like that
 *
 ************************************************************/


#define MAX_INSTANCES 4 /* max of 4 hardware devices loaded */

/* 
 * Note that this mutex gets used in _init. Which means this mutex
 * isn't just for this "instance" of the driver. It is for ALL INSTANCES.
 * If you have multiple cards, you have multiple instances of the driver,
 * with this single shared mutex across them all
 */
static kmutex_t   aipcam_global_mutex;

/* This is SORT OF like	int (*array)[] , except you NEVER USE THIS DIRECTLY.
 * You just use it with ddi_get_soft_state().
 * 
 *    "Global Softstate Handle" 
 */
void * aipcam_glob_soft_handle;

/* This gets filled out on a per-instance basis by aipcam_attach */
struct aipcam_softstate {
	dev_info_t *dip; /* you always need this cached */
	kmutex_t instance_mutex;
	kcondvar_t read_cv;
	
	int openflags; /* tell if we have been open()ed or not */

	usb_config_descr_t usb_config;
	int usb_interfacenum;
	usb_endpoint_descr_t endpoint[8];/* we have 'alt' endpoints */
	usb_interface_descr_t usb_interface_d;
	usb_pipe_handle_t controlpipe;
	usb_pipe_handle_t pipe_handle; /* normal isoc data pipe */
	int bridge_id; /* normally 0x511, or 0xa511 */

	/* This driver isnt exactly a STREAMs driver.
	 * However, the USB framework gives us the data in STREAMS type
	 * message blocks. So we kind of fake like we ARE
	 * a STREAMS driver.. but a lossy one.
	 * own buffer/queue
	 * "end" is where to add new data. "start" is where read()
	 * should take data from
	 */
	mblk_t *readqueue;
	int queuelen;

	/* probed info about the camera */
	/* See aipcam_camconfig() */
	int maxwidth;
	int maxheight;
	SENSORTYPE sensortype;

};



/*****************************************************************/
/*****************************************************************/
/*              Start of actual kernel routines here             */
/*****************************************************************/
/*****************************************************************/


/* just a little debugging tool */
static void printusberr(int errnum){
	switch(errnum){
		case USB_SUCCESS:
			printf(" NO error: USB_SUCCESS!\n");
			break;
		case USB_FAILURE:
			printf("  ERR=USB_FAILURE\n");
			break;
		case USB_NO_RESOURCES:
			printf("  ERR=USB_NO_RESOURCES\n");
			break;
		case USB_NO_BANDWIDTH:
			printf("  ERR=USB_NO_BANDWIDTH\n");
			break;
		case USB_PIPE_RESERVED:
			printf("  ERR=USB_PIPE_RESERVED\n");
			break;
		case USB_PIPE_ERROR:
			printf("  ERR=USB_PIPE_ERROR\n");
			break;
		case USB_PIPE_BUSY:
			printf("  ERR=USB_PIPE_BUSY\n");
			break;
		case USB_NO_FRAME_NUMBER:
			printf("  ERR=USB_NO_FRAME_NUMBER\n");
			break;
		case USB_INVALID_START_FRAME:
			printf("  ERR=USB_INVALID_START_FRAME\n");
			break;
		default:
			printf("  ERR=%d (unknown USB err)\n",errnum);
	}
}

void
aipcam_debugprint(va_alist)
va_dcl
{
	va_list args;
	char *fmt;
	if(debug_level==0){
		return;
	}

	va_start(args);
	fmt = va_arg(args, char *);
	/* DO NOT USE cmn_err here!!! it will not be interrupt-safe */
	vprintf(fmt, args);
	va_end(args);
}



/*Like the aipcam_global_mutex above, this routine is GLOBAL across all
 * instances.
 * It actually gets called when you add_drv your driver from the command
 * line, before  * it is hooked up to any specific hardware.
 * See aipcam_attach() below for instance-specific stuff.
 *
 * HOWEVER... Its often nice to pre-allocate "softstate" space
 * here, if you dont mind hardcoding the total number of possible
 * instances you can have. So we do that.
 * 
 */
int
_init(void)
{
       int     ret;

	cmn_err(CE_NOTE,"aipcam: _init called\n");
	/*
               * Initialize global mutex before mod_install'ing driver.
               * If mod_install() fails, must clean up mutex initialization
               */
	mutex_init(&aipcam_global_mutex, "aipcam: Global Mutex",
                     MUTEX_DRIVER, (void *)NULL);

	/* This is kind of like
	 * malloc(sizeof(struct aipcam_softstate *MAX_INSTANCES))
	 */
	ret=ddi_soft_state_init(&aipcam_glob_soft_handle,
				sizeof(struct aipcam_softstate),
			        MAX_INSTANCES);
	if(ret!=0){
		cmn_err(CE_WARN,"soft_state_init failed");
		return (ret);
	}
	
	if ((ret = mod_install(&modlinkage)) != 0) {
		mutex_destroy(&aipcam_global_mutex);
		ddi_soft_state_fini(&aipcam_glob_soft_handle);
		cmn_err(CE_WARN, "aipcam: _init: mod_install FAILED:\n");
		return DDI_FAILURE;
	}

	cmn_err(CE_NOTE,"aipcam: _init returning SUCCESS\n");
	return (DDI_SUCCESS);
}

int
_info(struct modinfo *modinfop)
{
	
	aipcam_debugprint("aipcam: _info called\n");
	return (mod_info(&modlinkage, modinfop));
}

/* This is the match to the _init() routine. Clean up stuff allocated
 * specifically in _init() here.
 * Stuff in attach() should be cleaned up in detach()
 */
int
_fini(void)
{
	int ret;

	aipcam_debugprint("aipcam: _fini called\n");
	
        /*
         * If mod_remove() is successful, we destroy our global mutex
         */
	if ((ret = mod_remove(&modlinkage)) == 0) {
		ddi_soft_state_fini(&aipcam_glob_soft_handle);
		mutex_destroy(&aipcam_global_mutex);
	}
	return (ret);
}

/************************************************************/
/**********************************************************************/
/*  Our read buffer handling section                                 */
/**********************************************************************/

/* return 0 on okay, 1 on fail.
 * handle whatever allocation the read queue stuff needs
 */
static int aipcam_readqueue_init(struct aipcam_softstate *softstate)
{
	softstate->queuelen=0;
	return 0;
}

/*
 * Returns internal pointers to "clear" the buffer.
 * Deallocates message blocks in queue.
 * Make sure you have the instance_mutex before calling this.
 */
static void readqueue_reset(struct aipcam_softstate *softstate){
	mblk_t *mptr,*nextptr;
	if(softstate->queuelen==0) return;

	mptr=softstate->readqueue;
	while(mptr!=NULL){
		nextptr=mptr->b_next;
		freemsg(mptr);
		mptr=nextptr;
	}
	softstate->queuelen=0;
}


/* You had better hold the instance mutex or something before calling this*/
static void aipcam_readqueue_finish(struct aipcam_softstate *softstate)
{
	readqueue_reset(softstate);
}

/*
 * Note that we have a circular queue, and we do NOT overwrite.
 * After queue is 'full', we drop old entries to make room
 * for the newer data.
 *
 * YOU MUST HAVE instance_mutex BEFORE CALLING!!!
 */
static void
add_to_readqueue(struct aipcam_softstate *softstate, mblk_t *mblkptr){
	mblk_t *parsequeue;
	if(mblkptr==NULL){
		printf("DEBUG: ERR: add_to_readqueue passed NULL\n");
		return;
	}
	if(mblkptr->b_next !=NULL){
		printf("DEBUG: ERR: add_to_readqueue passed mblk with non-NULL b_next\n");
		printf("     dropping it.\n");
/* XXX Maybe caller does this?
		freemsg(mblkptr);
*/
		return;
	}

	if(softstate->queuelen==0){
		if(softstate->readqueue!=NULL){
			printf("DEBUG: queuelen=0, but readqueue=%p\n",
			       softstate->readqueue);
		}
		softstate->readqueue=copymsg(mblkptr);
		softstate->queuelen=1;
		return;
	}

	if(softstate->queuelen == AIPTEK_QUEUE_DEPTH){
		/* gotta ditch older data first */
		parsequeue=softstate->readqueue;
		softstate->readqueue=parsequeue->b_next;
		freemsg(parsequeue); 
		softstate->queuelen -= 1;
	}

	/* find end of queue */
	parsequeue=softstate->readqueue;
	while(parsequeue->b_next!=NULL){
		parsequeue=parsequeue->b_next;
	}
	parsequeue->b_next=copymsg(mblkptr);
	softstate->queuelen += 1;


} /* end of add_to_readqueue() */

/* Get(remove) a mblk_t from the readqueue.
 * returns NULL if nothing in queue.
 *
 * You are responsible for calling freemsg() on the returned block
 * once you are done with it.
 *
 * You should hold the instance mutex before calling this
 */
static mblk_t *
getmsg_from_readqueue(struct aipcam_softstate *softstate){
	mblk_t *mblkptr;
	if(softstate->queuelen==0){
		return NULL;
	}
	mblkptr=softstate->readqueue;
	if(mblkptr==NULL){
		printf("ERR: getmsg_from_readqueue: queuelen=%d, but queue NULL\n",
		       softstate->queuelen);
		return NULL;
	}
	softstate->readqueue=mblkptr->b_next;
	softstate->queuelen -= 1;

	return mblkptr;
	
}

/**********************************************************************/

/* Utility routine. 
 * For each functional 'endpoint' for a USB device,there may be alternate
 * configurations to access it.
 * This routine tells the device to put an endpoint into a specific
 * alternate configuration
 *
 * Return USB_SUCCESS on okay.
 */
static int aipcam_set_altendpoint(struct aipcam_softstate *softstate,
				     int interfacenum, int altnum)
{
	uint_t failreason;
	int stat;
	
	
	stat=usb_pipe_sync_device_ctrl_send(softstate->controlpipe,
		USB_DEV_REQ_HOST_TO_DEV | USB_DEV_REQ_RECIPIENT_INTERFACE,
		USB_REQ_SET_INTERFACE,
		altnum,
		interfacenum,
		0,NULL, /* 0-len data */
		&failreason,
		USB_FLAGS_ENQUEUE);

	if(stat!=USB_SUCCESS){
		printf("DEBUG: set_altendpoint failed: reason=%d\n",
		       failreason);
	} else {
		aipcam_debugprint("Set endpoint to altnum=%d\n",altnum);
	}

	return stat;

}

/*
 * This is probably interrupt context. 
 */
static int
aipcam_isoc_callback(usb_pipe_handle_t pipe,
                        usb_opaque_t arg,
			usb_isoc_request_t *isoc_req)
{
	struct aipcam_softstate *softstate;
	mblk_t *mblkptr;
	int datalen;

	softstate=(struct aipcam_softstate*)arg;

	if(isoc_req==NULL){
		printf("isoc_callback: isoq_req==NULL\n");
		return (USB_SUCCESS);
	}

	if(isoc_req->isoc_data==NULL){
		aipcam_debugprint("isoc_callback: data ==NULL ? \n");
		usb_deallocate_isoc_req(pipe, isoc_req);
		return (USB_SUCCESS);
	} 

	if((isoc_req->isoc_pkt_descr[0].isoc_pkt_actual_length ==0x800) ||
	   (isoc_req->isoc_pkt_descr[0].isoc_pkt_actual_length ==0))
	{
		/* some stupid bug in sun handling of USB packets on x86.
		 * Max true packet len is 0x7ff. so 0x800 is 'special'
		 * 0x800 is a buggy flag for "This is an empty packet"
		 */
		printf("DEBUG: isoc_callback: empty packet. Throwing away\n");
		freemsg(isoc_req->isoc_data);
		usb_deallocate_isoc_req(pipe, isoc_req);
		return (USB_SUCCESS);
	}


	mblkptr=isoc_req->isoc_data;

	usb_deallocate_isoc_req(pipe, isoc_req);

	datalen=mblkptr->b_wptr - mblkptr->b_rptr;
	/* printf("Debug: isoc_callback got datalen=%d\n",datalen); */
		

	if(softstate->openflags != 0){
		mutex_enter(&softstate->instance_mutex);
		add_to_readqueue(softstate,mblkptr);
		cv_signal(&softstate->read_cv);
		mutex_exit(&softstate->instance_mutex);
	} else {
		printf("DEBUG: isoc_callback throwing away data: dev not open\n");
		/* does deallocate_isoc_req free the mblkptr?*/
	}

	return (USB_SUCCESS);
}
/* This gets called for partial USB transfers.
 * EG: some error occurred.
 */
static int
aipcam_isoc_exception_callback(usb_pipe_handle_t pipe, usb_opaque_t arg,
			 uint_t complete_reason, usb_isoc_request_t *isoc_req,
			 uint_t flag)
{
	/*struct aipcam_softstate *softstate=(struct aipcam_softstate*)arg;*/

	aipcam_debugprint("aipcam_EXCEPTION_callback called: reason %d\n",
			 complete_reason);
	return (USB_SUCCESS);
}

/* Although seemingly required for usb_pipe_start_polling(),
 * this routine seems to NEVER get called!
 */
static void
aipcam_poll_callback(usb_opaque_t arg,int arg1, uint_t arg2)
{
	aipcam_debugprint("ERROR: aipcam_poll_callback called with %x,%x\n");
	aipcam_debugprint("ASSUME that start_polling failed!\n");
	return;
}

static int
test_pipe_callback(usb_pipe_handle_t pipe, usb_opaque_t arg,mblk_t*mblkptr)
{
	printf("DEBUG: test_pipe_callback called\n");
	if(mblkptr!=NULL){
		freemsg(mblkptr);
	}
	return USB_SUCCESS;
}

static int
test_exception_callback(usb_pipe_handle_t pipe, usb_opaque_t arg,
			 uint_t complete_reason,mblk_t*mblkptr,uint_t flag)
{

	printf("DEBUG: test_EXCEPTION_callback called: reason %d\n",
			 complete_reason);
	if(mblkptr!=NULL){
		freemsg(mblkptr);
	}
	
	return (USB_SUCCESS);
}

/**********************************************************************/
/**********************************************************************/

/*      OmniVision (OV511) Chip specific routines      */


/*
 * write loveletters to the chip that controls the camera.
 * 0 on okay, 1 on fail
 */
static int aipcam_writereg(struct aipcam_softstate *softstate,
			  uint8_t reg, uint8_t value)

{
	mblk_t *mblkptr;
	int stat;
	uint_t failreason;

	mblkptr=allocb(1,0);
	*(mblkptr->b_wptr)=value&0xff;
	mblkptr->b_wptr++;
	

	stat=usb_pipe_sync_device_ctrl_send(softstate->controlpipe,
		USB_DEV_REQ_HOST_TO_DEV | USB_DEV_REQ_TYPE_VENDOR |
		  USB_DEV_REQ_RECIPIENT_DEVICE,
		2, /* bRequest*/
		   /* Normally, would be USB_REQ_XXX, but there isnt one
		    * for '2', which is what is supposedly needed here
		    */

		reg, /*wValue*/
		softstate->usb_interfacenum,
		1, /* wLength */
		mblkptr,
		&failreason,
		USB_FLAGS_ENQUEUE);

	if(stat != USB_SUCCESS){
		printf("ERR: reg_write failed: failreason=%d\n", failreason);
		printusberr(stat);
		return 1;
	}

	return 0;
}


/* read an OV511 register.
 * return 0 on okay, or 1 if something failed
 */
static int aipcam_readreg(struct aipcam_softstate *softstate,
			     uint8_t reg, uint8_t *value)

{
	mblk_t *mblkptr;
	int stat;
	uint_t failreason;

	printf("DEBUG: aipcam_readreg: using interfacenum %d\n",
		softstate->usb_interfacenum);

	stat=usb_pipe_sync_device_ctrl_receive(softstate->controlpipe,
		USB_DEV_REQ_DEVICE_TO_HOST | USB_DEV_REQ_TYPE_VENDOR |
		  USB_DEV_REQ_RECIPIENT_DEVICE,
		USB_REQ_SET_FEATURE, /* bRequest*/

		reg,                           /*wValue*/
		softstate->usb_interfacenum,   /*wIndex*/
		1,                             /* wLength */
		&mblkptr,                      /*data */
		&failreason,                   /*completion reason*/
		USB_FLAGS_ENQUEUE);            /* flags */

	if(stat != USB_SUCCESS){
		printf("ERR: reg_write failed: failreason=%d\n", failreason);
		printusberr(stat);
		return 1;
	}
	*value=*mblkptr->b_rptr;
/*	printf("aipcam_readreg: len of data is %d\n",
	       mblkptr->b_wptr - mblkptr->b_rptr);
*/

	freemsg(mblkptr);

	return 0;
}


/* wait for I2C "bus" on the other side of the USB bus to be idle.
 * return true (1) if clean idle.
 * return false (0) if either error on read, or device
 * flagged a NACK condition
 */
static int i2c_bus_idle(struct aipcam_softstate *softstate)
{
	int loop=0,wstatus;
	uint8_t i2c_status=0;

	/* Arrg. I hate busy-waits */
	/* linux code loops on this FOREVER! */
	while(++loop<10){
		wstatus=aipcam_readreg(softstate,
				       OV511_REG_I2C,
		                       &i2c_status);
		if(wstatus!=0){
			printf("DEBUG: i2c errored waiting for idle\n");
			return (0);
		}
		if(i2c_status & I2C_IDLE)
			break;
		drv_usecwait(1000);
	}

	if((i2c_status & I2C_NACK) ==0){
		return 1;
	} else {
		aipcam_debugprint("DEBUG: i2c_bus_idle: NACK encountered (stat=0x%x)\n",
				  i2c_status);
		return 0;
	}
}

/* do a kind of "port I/O" thing, to write through the USB bridge,
 * to the hardware on the other side of the I2C bus, in the camera.
 * return 0 on okay, 1 on fail.
 */
static int aipcam_writei2c(struct aipcam_softstate *softstate,
			      uint8_t reg, uint8_t value)
{
	int slave_writeid=0x42; /* OV7XX0 device */
	int wstatus=1; /* default to bad status */

	wstatus=aipcam_writereg(softstate,OV511_REG_I2C_SID,slave_writeid);
	if(wstatus==1) return 1;

	wstatus=aipcam_writereg(softstate,OV511_REG_I2C_SWA,reg);
	if(wstatus==1) return 1;

	wstatus=aipcam_writereg(softstate,OV511_REG_I2C_DATA,value);
	if(wstatus==1) return 1;

	wstatus=aipcam_writereg(softstate,OV511_REG_I2C,
					   I2C_3B_WT|I2C_START);

	if(wstatus!=0){
		aipcam_debugprint("DEBUG: i2c write failed\n");
	} else {
		aipcam_debugprint("DEBUG: i2c write(0x%x => 0x%x) succeeded\n",
				  value, reg);
	}
	return wstatus;
}


/* do a kind of "port I/O" thing, to read through the USB bridge,
 * to the hardware on the other side of the I2C bus, in the camera.
 * return 0 on okay, 1 on fail.
 */
static int aipcam_readi2c(struct aipcam_softstate *softstate,
			      uint8_t reg, uint8_t *valueptr)
{
	int wstatus=1;
	int slave_writeid=0x42;
	int slave_readid=0x42;

	/*reset cam first: write 0x80 to i2c reg 0x12 */
	wstatus=aipcam_writei2c(softstate,OV7XX0_COM_A, OV7XX0_COMA_RESET);
	if(wstatus!=0){
		printf("readi2c: initial COM_A reset failed\n");
		return 0;
	} else {
		printf("readi2c: wrote reset to COM_A okay\n");
	}
	/* 150ms wait */
	drv_usecwait(150000);

	/* This sequence is taken from "I2C R/W Sequence Examples",
	 * page 22 of the OV511 data sheet rev 1.0
	 */

	/* Note 100% sure this write ie nessessary, but.. */
	/*wstatus=aipcam_writereg(softstate,OV511_REG_I2C_SID,slave_writeid);
	if(wstatus==1) return 1;
	*/

	/* First, do "2 byte write" to set target register for read */
	wstatus=aipcam_writereg(softstate,OV511_REG_I2C_SMA,reg);
	if(wstatus==1){
		aipcam_debugprint(" readi2c failed write to SMA\n");
		return 1;
	}
	wstatus=aipcam_writereg(softstate,OV511_REG_I2C, I2C_2B_WT|I2C_START);
	if(wstatus==1){
		aipcam_debugprint(" readi2c failed I2C write start\n");
		return 1;
	}


	/* Now, actual "2 byte read", even though read 1 byte */
	wstatus=aipcam_writereg(softstate,OV511_REG_I2C_SRA,slave_readid);
	if(wstatus==1){
		aipcam_debugprint(" readi2c failed write to SRA\n");
		return 1;
	}
	wstatus=aipcam_writereg(softstate,OV511_REG_I2C,I2C_2B_RD|I2C_START);
	if(wstatus==1){
		aipcam_debugprint(" readi2c failed 2b-read start\n");
		return 1;
	}

	if(i2c_bus_idle(softstate)) {
		printf(" readi2c: passed bus_idle check\n");
	} else {
		printf(" readi2c: failed bus_idle check\n");
		return 1;
	}


	wstatus=aipcam_readreg(softstate,OV511_REG_I2C_DATA,valueptr);

	/* Kinda reset stuff either way. I guess.
	 * Doesnt make sense, but its just what you need to do.
	 * "clear bits by writing" or something, I guess.
	 */
	/*
	wstatus=aipcam_writereg(softstate,OV511_REG_I2C, I2C_2B_RD|I2C_START);
	*/

	return wstatus;


}

/* reset camera (sensor) hardware.
 * This is mainly a utility routine, rather than a "main" routine.
 *
 * This both sends a reset request, AND waits around and verifies
 * that everything comes out okay.
 *
 * Return 0 on okay, 1 on fail
 */
static int aipcam_resetcam(struct aipcam_softstate *softstate)
{
	int writestat;
	uint8_t reg1=0, reg2=0;

	aipcam_debugprint("aipcam_resetcam entered\n");

	writestat=aipcam_writei2c(softstate,OV7XX0_COM_A, 0x80);
	if(writestat!=0){
		return 1;
	} else {
		printf("aipcam_resetcam: written 0x80 okay\n");
	}
	/* Yes, do it steenkin AGAIN... */
	writestat=aipcam_writei2c(softstate,OV7XX0_COM_A, 0x80);
	if(writestat!=0){
		return 1;
	} else {
		printf("aipcam_resetcam: written 0x80 okay twice\n");
	}

	/* sensor takes time to reset itself. sigh. */
	/* Dunno how long. */
	drv_usecwait(200*1000);
	if(i2c_bus_idle(softstate)){
		printf("After usecwait, i2c bus IS idle\n");
	} else {
		printf("After usecwait, i2c bus is NOT idle\n");
	}

	/* apparently, a "dummy read" helps here */
#ifdef NONONN
	printf("DEBUG: resetcam doing dummy read \n");
	aipcam_readi2c(softstate,0,&reg1);
	printf("DEBUG: resetcam done dummy read \n");
#endif
	
	printf("DEBUG: resetcam reading i2c ID registers \n");
	aipcam_readi2c(softstate,OV7XX0_ID_MSB,&reg1);
	aipcam_readi2c(softstate,OV7XX0_ID_LSB,&reg2);
	printf("DEBUG: ID values are %x %x\n",reg1,reg2);
	if((reg1!=0x7f) || (reg2 != 0xa2)){
		printf(" ERROR: sensor ID values do not match\n");
		return 1;
	}
	printf(" DEBUG: resetcam says sensor IDs are okay\n");
	return 0;
	
}



/* Flexible reset. Used after changing any settings on the device.
 * Call with one of the OV511_RESET_XXX defines.
 * return 0 on okay, 1 on fail
 */
static int aipcam_resetbridge(struct aipcam_softstate *softstate, uint8_t rst){
	if(aipcam_writereg(softstate, OV511_REG_RESET,rst) != 0){
		printf("DEBUG: reset write failed\n");
		return 1;
	}
	if(aipcam_writereg(softstate, OV511_REG_RESET, 0) != 0) {
		printf("DEBUG: reset write of 0 failed\n");
		return 1;
	}
	return 0;
}

/*
 * set I2C "slave IDs".
 * dont see why we should have to do this. ARRGGG...
 * return 1 on bad, 0 on fail
 */
static int set_i2c_slaveids(struct aipcam_softstate *softstate)
{
	/*assume OV7xx device for now. dammit */
	int slaveid=0x42; /* OV7XX0 device */
	int writestat=0;
	uint8_t reg1=0,reg2=0;

	aipcam_debugprint("set_i2c_slaveids called\n");

	/* DEBUG bit */
	aipcam_readreg(softstate,OV511_REG_I2C_SID, &reg1);
	aipcam_readreg(softstate,OV511_REG_I2C_SRA, &reg2);
	printf("DEBUG: initial IDs are %x,%x: will set to 0x42,0x43\n",
	       reg1,reg2);

	/* first set slaveid. Then see if it works, by trying to
	 * write to a known register
	 */
	writestat=aipcam_writereg(softstate,OV511_REG_I2C_SID, slaveid);
	writestat+=aipcam_writereg(softstate,OV511_REG_I2C_SRA, slaveid+1);
	if(writestat!=0){
		printf("DEBUG: could not set i2c slave IDs\n");
		return 1;
	}

	/* attempt to make things happy, suggested by Mark */
	writestat=aipcam_writereg(softstate,OV511_REG_I2C_CLOCK, 0x3);
	if(writestat!=0){
		printf("DEBUG: could not set i2c clock\n");
		return 1;
	} else {
		printf("Set I2C Clock to 3 successfully\n\n");
	}
	

	writestat=aipcam_resetbridge(softstate,OV511_RESET_MOST);
	if(writestat!=0){
		printf("DEBUG: set_i2c_slaveids: reset failed\n");
		return 1;
	}
	drv_usecwait(drv_usectohz(150000)); /* wait 150milisec */

	aipcam_readreg(softstate,OV511_REG_I2C_SID, &reg1);
	aipcam_readreg(softstate,OV511_REG_I2C_SRA, &reg2);
	printf("DEBUG: set_i2c_slave: after reset, IDs are now %x,%x\n",
	       reg1,reg2);

	return writestat;
}

static void set_resolution(struct aipcam_softstate *softstate)
{
	int pxcount=0x4f;
	int lncount=0x3d;
	int stat=0;
	stat+=aipcam_writereg(softstate,OV511_REG_PXCNT,pxcount);
	stat+=aipcam_writereg(softstate,OV511_REG_LNCNT,lncount);
	stat+=aipcam_writereg(softstate,OV511_REG_PXDIV,0);
	stat+=aipcam_writereg(softstate,OV511_REG_LNDIV,0);
	if(stat==0){
		printf("DEBUG: set_resolution1 succeeded\n");
	}else {
		printf("DEBUG: set_resolution1 failed\n");
	}

	stat=0;
	stat+=aipcam_writereg(softstate,OV511_REG_SNAP_PXCNT,pxcount);
	stat+=aipcam_writereg(softstate,OV511_REG_SNAP_LNCNT,lncount);
	stat+=aipcam_writereg(softstate,OV511_REG_SNAP_PXDIV,0);
	stat+=aipcam_writereg(softstate,OV511_REG_SNAP_LNDIV,0);

	if(stat==0){
		printf("DEBUG: set_resolution succeeded\n");
	}else {
		printf("DEBUG: set_resolution failed\n");
	}
}

/* Initial configuration of USB "bridge" hardware (OV511)
 * Not to be confused with configuring the CAMERA hardware.
 * Return 1 on fail, 0 on good.
 */

static int aipcam_init511(struct aipcam_softstate *softstate)
{
	/* mystical sequence handed down to me by Mark McClelland */
	if(aipcam_writereg(softstate, OV511_REG_RESET, OV511_RESET_ALL)){
		return 1;
	}
	if(aipcam_writereg(softstate, OV511_REG_ENSYS, 1)) {
		return 1;
	}
	if(aipcam_writereg(softstate, OV511_REG_RESET, OV511_RESET_ALL)){
		return 1;
	}
	if(aipcam_writereg(softstate, OV511_REG_ENSYS, 1)) {
		return 1;
	}
	if(aipcam_writereg(softstate, OV511_REG_RESET, OV511_RESET_MOST)){
		return 1;
	}
	if(aipcam_writereg(softstate, OV511_REG_ENSYS, 1)) {
		return 1;
	}

	if(aipcam_writereg(softstate, OV511_REG_RESET, OV511_RESET_STOPCAM)){
		return 1;
	}

	/* part 1 done! We have reset the hardware*/
	/* we might take this brief pause to fiddle with the LED */
	/* Now on to part 2: normal configuration */

	/* Another Bizzare, magical sequence */
	if(aipcam_writereg(softstate, OV511_REG_ENFC, OV511_ENFC_ON)){
		return 1;
	}
	if(aipcam_writereg(softstate, OV511_REG_SNAP, 0x01)){
		return 1;
	}
	if(aipcam_writereg(softstate, OV511_REG_SNAP, 0x03)){
		return 1;
	}
	if(aipcam_writereg(softstate, OV511_REG_SNAP, 0x01)){
		return 1;
	}

	if(aipcam_writereg(softstate, OV511_REG_FIFO_OPTS, 0x1f)){
		return 1;
	}
	/* turn off compression */
	if(aipcam_writereg(softstate, OV511_REG_CMP_EN, 0x0)){
		return 1;
	}

	/* enable color lookup  */
	if(aipcam_writereg(softstate, OV511_REG_LUT_EN,
			      OV511_LUT_Y_ENABLE|OV511_LUT_UV_ENABLE)){
		return 1;
	}

	return 0;
}


/* Initial configuration of USB "bridge" hardware (OV511A)
 * Not to be confused with configuring the CAMERA hardware.
 * Return 1 on fail, 0 on good.
 */

static int aipcam_init511a(struct aipcam_softstate *softstate)
{
	/* mystical sequence handed down to me by Mark McClelland */
	if(aipcam_writereg(softstate, OV511_REG_RESET, OV511_RESET_ALL)){
		return 1;
	}
	if(aipcam_writereg(softstate, OV511_REG_ENSYS, 1)) {
		return 1;
	}
	if(aipcam_writereg(softstate, OV511_REG_RESET, OV511_RESET_ALL)){
		return 1;
	}
	if(aipcam_writereg(softstate, OV511_REG_ENSYS, 1)) {
		return 1;
	}
	if(aipcam_writereg(softstate, OV511_REG_RESET, OV511_RESET_MOST)){
		return 1;
	}
	if(aipcam_writereg(softstate, OV511_REG_ENSYS, 1)) {
		return 1;
	}

	if(aipcam_writereg(softstate, OV511_REG_RESET, OV511_RESET_STOPCAM)){
		return 1;
	}

	/* part 1 done! We have reset the hardware*/
	/* we might take this brief pause to fiddle with the LED */
	/* Now on to part 2: normal configuration */

	/* Another Bizzare, magical sequence */
	if(aipcam_writereg(softstate, OV511_REG_ENFC, 0xff)){
		return 1;
	}
	if(aipcam_writereg(softstate, OV511_REG_SNAP, 0x01)){
		return 1;
	}
	if(aipcam_writereg(softstate, OV511_REG_SNAP, 0x03)){
		return 1;
	}
	if(aipcam_writereg(softstate, OV511_REG_SNAP, 0x01)){
		return 1;
	}

	if(aipcam_writereg(softstate, OV511_REG_FIFO_OPTS, 0xff)){
		return 1;
	}
	/* turn off compression */
	if(aipcam_writereg(softstate, OV511_REG_CMP_EN, 0x0)){
		return 1;
	}

	/* enable color lookup  */
	if(aipcam_writereg(softstate, OV511_REG_LUT_EN,
			      OV511_LUT_Y_ENABLE|OV511_LUT_UV_ENABLE)){
		return 1;
	}

	return 0;
}

/* Just verify that we can read the OV511 'bridge' hardware */
static void check_bridge(struct aipcam_softstate *softstate)
{
	uint8_t regval=0, newval=0xfd;
	aipcam_readreg(softstate,OV511_REG_USERREG,&regval);
	printf("Checking USER reg.  Initially = 0x%x\n",regval);

	aipcam_writereg(softstate,OV511_REG_USERREG,newval);
	aipcam_readreg(softstate,OV511_REG_USERREG,&regval);
	if(regval==newval){
		printf("Debug: USER reg now correctly shows 0x%x\n",regval);
	} else {
		printf("Debug: USER reg shows 0x%x instead of 0x%x\n", regval,newval);
	}
	printf("  Resetting it to 0\n");
	aipcam_writereg(softstate,OV511_REG_USERREG,0x0);
	aipcam_readreg(softstate,OV511_REG_USERREG,&regval);
	printf("Debug: USER reg now correctly shows 0x%x\n",regval);
}


/* TEMPORARY debug thing.
 * Scan things as best we can, and report on them.
 * return 1 on good, 0 on bad
 */
static int identify_hardware(struct aipcam_softstate *softstate)
{
	int statval;
	uint8_t reg;

	printf("DEBUG: identify_hardware called\n");

	statval=aipcam_readreg(softstate,0x1a,&reg);
	printf("USB reg at 1a should be 0x4f: is %0x\n",reg);
	statval=aipcam_readreg(softstate,0x1b,&reg);
	printf("USB reg at 1b should be 0x1d: is %0x\n",reg);
	statval=aipcam_readreg(softstate,0x23,&reg);
	printf("USB reg at 23 should be 0x1a: is %0x\n",reg);

	statval=aipcam_readreg(softstate,OV511_REG_CUST_ID,&reg);
	if(statval!=0){
		printf("ERR: read of CUST_ID reg failed\n");
		return 0;
	} else {
		printf("DEBUG: USB bridge CUST_ID=%d\n",reg);
	}

	if(set_i2c_slaveids(softstate)!=0){
		printf("DEBUG: set_slaveid in identify returned fail\n");
		return 0;
	} else {
		printf("DEBUG: set_slaveid in identify worked\n");
	}

	/*write 0x80 to i2c reg 0x12: reset cam */
	statval=aipcam_writei2c(softstate,OV7XX0_COM_A, OV7XX0_COMA_RESET);
	if(statval!=0){
		printf("DEBUG: identify: write to COMA failed\n");
		return 0;
	} else {
		printf("identify: wrote reset to COM_A okay\n");
	}
	/* 150ms wait */
	drv_usecwait(150000);

	statval=aipcam_writei2c(softstate,OV7XX0_COM_A, 0x80);
	if(statval!=0){
		printf("DEBUG: identify: SECOND write to COMA failed\n");
		return 0;
	} else {
		printf("identify: wrote COM_A AGAIN\n");
	}
	
	/* 150ms wait */
	drv_usecwait(150000);

/*
	statval=aipcam_readi2c(softstate,OV7XX0_COM_I,&reg);
	if(statval!=0){
		printf("read of COM_I failed\n");
		return 0;
	} else {
		printf("identify: COM_I is %x\n",reg);
	}

*/

	printf("DEBUG: Going for read of OV7XX0_ID_MSB\n");

	reg=0;
	statval=aipcam_readi2c(softstate,OV7XX0_ID_MSB,&reg);
	if(statval!=0){
		printf("read of MSB failed\n");
		return 0;
	} else {
		printf("identify: MSB is %x\n",reg);
	}
	reg=0;
	statval=aipcam_readi2c(softstate,OV7XX0_ID_LSB,&reg);
	if(statval!=0){
		printf("read of LSB failed\n");
		return 0;
	} else {
		printf("identify: LSB is %x\n",reg);
	}

	return 1;

}

/* configures actual camera hardware, as opposed to USB stuff.
 * Not to be confused with, or substituted for, intialization of
 * the USB "bridge" hardware. see aipcam_initbridge()
 * According to Mark McClelland's drivers, this should work for
 * sensor types OV7610, OV7620, and UV7620AE
 * Similar to his ov7xx0_configure() routine
 *
 * Return 0 on okay, 1 on fail.
 */
static int aipcam_initcam(struct aipcam_softstate *softstate)
{
	uint8_t regval;

	softstate->sensortype=UNKNOWN;

	aipcam_debugprint("initcam entered\n");

#ifdef DO_RESET
	if(aipcam_resetcam(softstate)==1){
		printf("DEBUG: initcam: i2c reset FAILED\n");
		return 1;
	}
	printf("DEBUG: initcam: i2c reset returned OK.\n");
#else
	printf("DEBUG: initcam:  SKIPPED resetcam.\n");
#endif
       
	/* Now try to figure out what kind of sensor hardware I have */
	if(aipcam_readi2c(softstate,OV7XX0_COM_I, &regval) !=0){
		return 1;
	}
	if((regval & 3)==3){
		aipcam_debugprint("Cam sensor is OV7610");
		softstate->sensortype=OV7610;
	} else if((regval & 3)==0){
		aipcam_debugprint("Cam sensor is OV7620");
		softstate->sensortype=OV7620;
	} else {
		printf("aipcam: Got one of them wierd sensor types: 0x%x\n",
		       regval);
		return 1;
	}

	softstate->maxwidth=640;
	softstate->maxheight=480;
	return 0;
}

/*
 * Close existing pipe, if one is open.
 * Then open one of the 8 possible 'alt' endpoint variants, depending on
 * desired package size.
 * We will also set the hardware to the appropriate settings first.
 *
 * return 0 on okay, 1 on fail.
 */
static int aipcam_open_endpoint(struct aipcam_softstate *softstate,
				   int eptnum)
{
	/* OV511:     1=993, 3=769, 5=513, 6=257 */
	/* OV511plus: 1=33, 2=129, 3=257, 4=385, 5=513, 6=769,7=961 */

	int endpoint_alt, stat;
	uint8_t fifosize;
	usb_pipe_policy_t policy,*policyptr;
	int maxpktsize;

	printf("DEBUG: aipcam_open_endpoint called\n");

	if(aipcam_writereg(softstate, OV511_REG_RESET, 0x0)){
		return 1;
	}

	if(softstate->pipe_handle!=NULL){
		printf("DEBUG: already have a pipe open?? closing it.\n");
		usb_pipe_stop_polling(softstate->pipe_handle,
			      USB_FLAGS_SLEEP,NULL,NULL);
		usb_pipe_close(&softstate->pipe_handle,
			  USB_FLAGS_SLEEP,NULL,NULL );
		softstate->pipe_handle=NULL;
	}

	/* only for plain 511 right now */
	switch(eptnum){
		case AIPTEK_ENDPOINT_993:
			endpoint_alt=1;
			fifosize=0x1f;
			break;
	/*	case AIPTEK_ENDPOINT_769:
			endpoint_alt=3;
			break;
		case AIPTEK_ENDPOINT_513:
			endpoint_alt=5;
			break;
		case AIPTEK_ENDPOINT_257:
			endpoint_alt=6;
			break;
	*/
		default:
			printf("DEBUG: unsupported endpointspeed num %d\n",
			       eptnum);
			return 1;
	}

	
	stat=aipcam_writereg(softstate,OV511_REG_FIFOSIZE,fifosize);
	if(stat!=0){
		cmn_err(CE_WARN,"Error setting FIFO size");
		return 1;
	}

	stat=aipcam_writereg(softstate,OV511_REG_RESET,
				 OV511_RESET_STOPCAM);
	stat+=aipcam_writereg(softstate,OV511_REG_RESET,0);

	if(stat!=0){
		cmn_err(CE_WARN,"Error in post-FIFO reset");
		return 1;
	}


	set_resolution(softstate);

	/* cant believe we have to do this by hand, as WELL as
	 * using the damn separate endpoint thingy..
	 */

	stat=aipcam_set_altendpoint(softstate,softstate->usb_interfacenum,
				       endpoint_alt);

	if(stat!=USB_SUCCESS){
		cmn_err(CE_WARN,"set_altendpoint returned fail");
		printusberr(stat);
		mutex_exit(&softstate->instance_mutex);
		return (ENXIO);
	}
	

	maxpktsize=softstate->endpoint[endpoint_alt].wMaxPacketSize;
	printf("DEBUG: maxpkgsize of endpoint=%d\n",maxpktsize);

	policyptr=&policy;
	bzero(policyptr,sizeof(policy));
	policyptr->pp_version=USB_PIPE_POLICY_V_0;
	policyptr->pp_timeout_value=USB_PIPE_TIMEOUT;
	/* defaults to wMaxPacketSize, supposedly */
	/*policyptr->pp_periodic_max_transfer_size=maxpktsize;*/

	policyptr->pp_isoc_rcv_pkt_size=maxpktsize;
	policyptr->pp_isoc_rcv_pkts_per_callback=1;

	policyptr->pp_isoc_callback=aipcam_isoc_callback;
	policyptr->pp_isoc_exception_callback=aipcam_isoc_exception_callback;
	policyptr->pp_callback_arg=(void*)softstate;/* same for ALL callbacks*/

	policyptr->pp_callback=test_pipe_callback;
	policyptr->pp_exception_callback=test_exception_callback;


	stat=usb_pipe_open(softstate->dip, &softstate->endpoint[endpoint_alt],
			   policyptr, USB_FLAGS_SLEEP,
			   &softstate->pipe_handle);

	if(stat !=USB_SUCCESS){
		cmn_err(CE_WARN,"usb_pipe_open failed: ");
		printusberr(stat);
		mutex_exit(&softstate->instance_mutex);
		return (ENXIO);
	} else {
		printf("DEBUG: open_endpoint succeeded opening pipe for alt=%d\n",
		       endpoint_alt);
	}

	return 0;
}


/**********************************************************************/
/**********************************************************************/


/* Kinda wierd, but this is an official solaris entry point, that can
 * be registered by cb_ops. We dont use it, but we provide it
 * for the OS to use.
 */
static int aipcam_print(dev_t dev, char*string){
	int instance=DEVT_TO_INST(dev);
	cmn_err(CE_NOTE, "base-print(%d): %s",instance,string);
	return 0;
}


static int aipcam_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd,
			     void *arg, void **result){
	struct aipcam_softstate *softstate;
	int returnval=DDI_FAILURE;
	int instance=getminor((dev_t)arg);
	
	aipcam_debugprint("aipcam: getinfo called\n");
	
	switch(infocmd){
	    case DDI_INFO_DEVT2DEVINFO:
		softstate=ddi_get_soft_state(aipcam_glob_soft_handle,instance);
		if(softstate==NULL){
			*result=NULL;
			break;
		}
		*result=*softstate->dip;
		returnval=DDI_SUCCESS;
		break;
	    case DDI_INFO_DEVT2INSTANCE:
		*result=(void*)instance;
		returnval=DDI_SUCCESS;
		break;
	}

	return (returnval);
}




/* you probably wont even need to have a probe for a PCI driver 
 * But just in case you want one to tell you what major and minor
 * you're being probed with, here's one
 */
 
static int aipcam_probe(dev_info_t *dip){
#ifndef DO_PROBE_ROUTINE
   	aipcam_debugprint("aipcam: probe called\n");
#else

  	target=ddi_prop_get_int(DDI_DEV_T_ANY,dip,DDI_PROP_DONTPASS,
				"target", -1);
  	lun=ddi_prop_get_int(DDI_DEV_T_ANY,dip,DDI_PROP_DONTPASS,
				"lun", -1);
   
	aipcam_debugprint("probe- called fortarget=%d, lun=%d\n", target,lun);

	/* Normally, we would check the dev_info info,
	 * to see which device ID is being probed, and then
	 * check to see if the durn thing is actually there.
	 * But for now, assume .conf file is correct
	 */
   
	/* our two options for success are:
	 * DDI_PROBE_DONTCARE
	 * DDI_PROBE_SUCCESS
	 */

#endif /* DO_PROBE_ROUTINE */

	return DDI_PROBE_DONTCARE;
	/*return DDI_PROBE_FAILURE;*/
}

/* This routine can be called by aipcam_attach (if errors occur)
 * and aipcam_detach (on normal detach)
 * Idea is to clean up any mutexes, register mappings, etc. nicely.
 * We also free the instance-softstate.
 * We do NOT remove the minor node, however. Only detach does
 * that currently.
 */
static void aipcam_cleanup(struct aipcam_softstate *softstate){
	int instance;

	if(softstate==NULL){
		/* yes, paranoia is good*/
		cmn_err(CE_WARN,"ERROR: aipcam_cleanup called with NULL");
		return;
	}

	if(softstate->pipe_handle!=NULL){
		usb_pipe_stop_polling(softstate->pipe_handle,
			      USB_FLAGS_SLEEP,NULL,NULL);
		usb_pipe_close(&softstate->pipe_handle,
			  USB_FLAGS_SLEEP,NULL,NULL );
		softstate->pipe_handle=NULL;
	}
	if(softstate->controlpipe!=NULL){
		usb_pipe_close(&softstate->controlpipe,
			  USB_FLAGS_SLEEP,NULL,NULL );
		softstate->controlpipe=NULL;
	}

	aipcam_readqueue_finish(softstate);

   	instance=ddi_get_instance(softstate->dip);

	ddi_soft_state_free(aipcam_glob_soft_handle,instance);
}


/*
 * _init() has already been called. This routine gets called
 * once for EVERY MATCHING PIECE OF HARDWARE.
 * It gets called for initial "attach", and also for resume(/suspend)
 * You would usually allocate space for state variables, and
 * map device registers in here.
 *
 * To avoid reboot loops, be very very careful putting things in here.
 * If you want to put extra debug type stuff in, it's safer putting
 * it in aipcam_open() or someplace like that, that gets called after
 * the driver is safely loaded
 *
 * Be sure to free/deallocate anything you create here, in
 *  aipcam_detach()
 */
static int aipcam_attach(dev_info_t *dip, ddi_attach_cmd_t cmd){
   	int instance,stat;
	struct aipcam_softstate *softstate;
	size_t usb_raw_len;
	uchar_t *usb_rawconf;
	usb_device_descr_t *dev_desc;

	cmn_err(CE_NOTE,"aipcam: attach called\n");

	/* default debug level to **0**
	 * You can override in the .conf file.
	 */
	debug_level=ddi_prop_get_int(DDI_DEV_T_ANY,dip, DDI_PROP_DONTPASS,
                                    "debug-level", 1);

	aipcam_debugprint("aipcam: aipcam_attach called\n");

   	switch(cmd){
	 case DDI_RESUME:
	   	cmn_err(CE_NOTE, "aipcam: Ignoring attach_RESUME\n");
		return DDI_FAILURE;
	 case DDI_PM_RESUME:
		cmn_err(CE_NOTE, "aipcam: Ignoring attach_PM_RESUME\n");
		return DDI_FAILURE;
	 case DDI_ATTACH:
	   	break;
	 default:
	   	cmn_err(CE_NOTE, "aipcam: unknown attach command\n");
	   	return DDI_FAILURE;
	}
   	instance=ddi_get_instance(dip);
   	aipcam_debugprint("aipcam: attach for instance %d\n", instance);


	/* We dont have to check for being over MAX_INSTANCES, because
	 * these two calls do it implicitly
	 */
	if (ddi_soft_state_zalloc(aipcam_glob_soft_handle, instance) != DDI_SUCCESS) {
		cmn_err(CE_WARN, "aipcam: %d: Failed to alloc soft state",
		    instance);
		return (DDI_FAILURE);
	}
	softstate=ddi_get_soft_state(aipcam_glob_soft_handle,instance);
	if(softstate==NULL){
		cmn_err(CE_WARN, "aipcam: attach for instance %d: Failed to get soft state",
			instance);
		ddi_soft_state_free(aipcam_glob_soft_handle,instance);
		return (DDI_FAILURE);
	}

	softstate->dip = dip;
	softstate->openflags = 0; /* redundant, but what the heck */
	
	usb_rawconf = usb_get_raw_config_data(dip,&usb_raw_len);

	/* Fill out "configuration" descriptor */
	if(usb_parse_configuration_descr(usb_rawconf, usb_raw_len,
					&softstate->usb_config,
					USB_CONF_DESCR_SIZE) !=
						USB_CONF_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"usb_parse_configuration_desc failed\n");
		aipcam_cleanup(softstate);
		return DDI_FAILURE;
	}


	aipcam_debugprint("USB config num is %d\n",
			 softstate->usb_config.bConfigurationValue);
	softstate->usb_interfacenum=usb_get_interface_number(dip);
	aipcam_debugprint("USB interface num is %d\n",
				softstate->usb_interfacenum);

	/* If interfacenum==-1, set it to 0: we are responsible for full dev*/
	if(softstate->usb_interfacenum==-1){
		aipcam_debugprint("  (resetting interface num to 0)\n");
		softstate->usb_interfacenum=0;
	}

	/* Fill out "interface" descriptor */
	if(usb_parse_interface_descr(usb_rawconf,usb_raw_len,
				    softstate->usb_interfacenum,
				    0,
				    &softstate->usb_interface_d,
				    USB_IF_DESCR_SIZE) != USB_IF_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"usb_parse_interface_desc failed");
		aipcam_cleanup(softstate);
		return DDI_FAILURE;
	}

	dev_desc=usb_get_dev_descr(dip);
	aipcam_debugprint("DEBUG: idProduct is %d(0x%x)\n",
	       dev_desc->idProduct,dev_desc->idProduct);

	softstate->bridge_id=dev_desc->idProduct;
	switch(softstate->bridge_id){
		case 0x511:
		case 0xA511:
			break;
		default:
			printf("Sorry: %x not a supported videocam \n",
			       softstate->bridge_id);
			aipcam_cleanup(softstate);
			return DDI_FAILURE;
			break;
	}

#ifdef PRINTSTRINGS
	{
		char stringbuf[100];

		usb_get_string_descriptor(dip,USB_LANG_ID,
					  softstate->usb_interface_d.iInterface,
					  stringbuf,15);
		printf("DEBUG: manu=%s\n",stringbuf);
		usb_get_string_descriptor(dip,USB_LANG_ID,
					  dev_desc->iProduct,
					  stringbuf,15);
		printf("DEBUG: product=%s\n",stringbuf);
		usb_get_string_descriptor(dip,USB_LANG_ID,
					  dev_desc->iProduct,
					  stringbuf,15);
		printf("DEBUG: interface desc=%s\n",stringbuf);
	}
#endif /* PRINTSTRINGS */
	aipcam_debugprint("number of endpoints = %d\n",
			 softstate->usb_interface_d.bNumEndpoints);


	/* initialize info for isoc-IN endpoint, in all its
	 * permutations
	 */
	{
		int altnum;
		for(altnum=0;altnum<8;altnum++){
			if(usb_parse_endpoint_descr(usb_rawconf,
					    usb_raw_len,
					    softstate->usb_interfacenum,
					    altnum, /* alt interfacenum*/
					    0, /*endpoint index num*/
					    &softstate->endpoint[altnum],
					    USB_EPT_DESCR_SIZE) !=
						USB_EPT_DESCR_SIZE)
			{
			cmn_err(CE_WARN, "could not parse endpoint %d:%d",
					0,altnum);
			return DDI_FAILURE;
			}
		}
	}

	stat=usb_pipe_open(softstate->dip,
	                 NULL, NULL,
	                 USB_FLAGS_OPEN_EXCL, 
	                 &softstate->controlpipe);
	if(stat != USB_SUCCESS) {
		printf("ERROR: could not open default pipe.\n");
		printusberr(stat);
		return DDI_FAILURE;
	}
			

	mutex_init(&softstate->instance_mutex, NULL, MUTEX_DRIVER, NULL);
	cv_init(&softstate->read_cv, NULL, CV_DRIVER,NULL);

	if(aipcam_readqueue_init(softstate)!=0){
		return DDI_FAILURE;
	}

	/* See if bridge hardware functions like we expect */
	check_bridge(softstate);

	switch(softstate->bridge_id){
		case 0x511:
			if(aipcam_init511(softstate) !=0){
				printf("Bridge Init of OV511 FAILED\n");
				return (DDI_FAILURE);
			}
			break;
		case 0xA511:
			if(aipcam_init511a(softstate) !=0){
				printf("Bridge Init of OV511+ FAILED\n");
				return (DDI_FAILURE);
			}
			break;
	}
	printf("initbridge returned OK\n");

	/* I suspect somehow this is screwing me up */

	if(set_i2c_slaveids(softstate)!=0){
		printf("DEBUG: set_slaveid returned fail\n");
		return DDI_FAILURE;
	} else {
		printf("DEBUG: set_slaveid worked\n");
	}

	if(!identify_hardware(softstate)){
		printf("ERROR: identify_hardware returned FAIL\n");
		return DDI_FAILURE;
	} else {
		printf("DEBUG: identify_hardware returned okay\n");
	}



	check_bridge(softstate);

	if(aipcam_initcam(softstate) !=0){
		printf("DEBUG: initcam FAILED\n");
	} else {
		printf("DEBUG: initcam returned OK\n");
	}

	ddi_create_minor_node(dip, DEVNODE_NAME, S_IFCHR,
			  INST_TO_NODENUM(instance),
			  DEVNODE_TYPE, 0);

	/* This little call all by itself, will put a whole bunch of
	 * useful information about the device in
	 * /var/adm/messages, aka "dmesg |tail"
	 */
	ddi_report_dev(dip);

	/* Dont forget to make ddi_remove_minor_node match in detach() */

	return DDI_SUCCESS;

}

/* used by read() to figure out how many bytes of data
 * the user program wants right now
 */
static int count_uio_size(struct uio* uiostruct){
	int datalen=0;
	int vec_count;
	if(uiostruct==NULL){
		return 0;
	}
	for(vec_count=0;vec_count<uiostruct->uio_iovcnt;vec_count++){
		datalen+=uiostruct->uio_iov[vec_count].iov_len;
	}

	return datalen;
}


/* return 0 on success, or errno */
static int aipcam_read(dev_t dev, struct uio* ustr, cred_t * credp){
	int instance=DEVT_TO_INST(dev);
	struct aipcam_softstate *softstate;
	int readbytes,msgbytes;
	mblk_t *mblkptr;

	aipcam_debugprint("aipcam: read called\n");

	softstate=ddi_get_soft_state(aipcam_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"aipcam_read failed: cannot get softstate?");
		return DDI_FAILURE;
	}
	
	mutex_enter(&softstate->instance_mutex);

	if(cv_wait_sig(&softstate->read_cv, &softstate->instance_mutex)<=0){
		printf("DEBUG: aipcam_read: signal received\n");
		mutex_exit(&softstate->instance_mutex);
		return EPIPE;
	}

	aipcam_debugprint("Supposedly, we have received read data\n");

	mblkptr=getmsg_from_readqueue(softstate);

	mutex_exit(&softstate->instance_mutex);

	if(mblkptr==NULL){
		printf("ERR: read() cv_signaled, but no data?\n");
		return ENODATA;
	}

	/* If user reads less than a whole block.. too bad.
	 * The block goes away anyway
	 * Which means you dont want to use FILE* on this device!
	 * Unless you set the buffer size to be an exact multiple
	 * of the datablock size, that is.
	 */
	readbytes=count_uio_size(ustr);
	msgbytes=mblkptr->b_wptr - mblkptr->b_rptr;
	if(msgbytes>readbytes) msgbytes=readbytes;

	if(uiomove(mblkptr->b_rptr, msgbytes,UIO_READ,ustr) !=0){
		aipcam_debugprint("ERR: uiomove failed\n");
		return EFAULT;
	}

	/* uiomove has already adjusted bytecount read in the uio struct */
	/* So we dont have to do anything else with it */

	freemsg(mblkptr);
	
	
   	return (0);

}

static int aipcam_write(dev_t dev, struct uio *uiop, cred_t *credp){
	int instance=DEVT_TO_INST(dev);
	struct aipcam_softstate *softstate;

   	aipcam_debugprint("aipcam: write called (huh?)\n");

	softstate=ddi_get_soft_state(aipcam_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"aipcam_write failed: cannot get softstate?");
		return DDI_FAILURE;
	}

   	return (ENXIO);
}


static int aipcam_open(dev_t *dev, int openflags, int otyp, cred_t *credp){
	int instance=DEVT_TO_INST(*dev);
	struct aipcam_softstate *softstate;
	usb_pipe_state_t pipestate;
	int stat;

	aipcam_debugprint("aipcam: open for instance %d\n",  instance);



	softstate=ddi_get_soft_state(aipcam_glob_soft_handle, instance);
	if(softstate==NULL){
		/* This is NOT technically a failure: it is sometimes expected.
		 * It is called "deferred attach"
		 */
		return (ENXIO);
	}

	mutex_enter(&softstate->instance_mutex);

	aipcam_debugprint("aipcam: openflags=%d\n",  softstate->openflags);
	if(softstate->openflags!=0){
		aipcam_debugprint("aipcam: device already opened\n");
		mutex_exit(&softstate->instance_mutex);
		return (EBUSY);
	}



	/* Clear out read queue. Probably redundant.*/
	readqueue_reset(softstate);

	if(aipcam_open_endpoint(softstate, AIPTEK_ENDPOINT_993)!=0){
		printf("DEBUG: open_endpoint failed\n");
		mutex_exit(&softstate->instance_mutex);
		return (ENXIO);
	} else {
		printf("DEBUG: open_endpoint returned OK\n");
	}

	softstate->openflags=1;

	if(usb_pipe_get_state(softstate->pipe_handle,
			      &pipestate,
			      USB_FLAGS_SLEEP) != USB_SUCCESS)
	{
		cmn_err(CE_WARN,"ERROR: could not get usb_pipe state!");
	}
	printf("DEBUG: pipe-state after open is %d\n",pipestate);
	if(pipestate!=USB_PIPE_STATE_IDLE){
		printf("DEBUG: ERROR: pipe should be IDLE\n");
		/*
		usb_pipe_reset(softstate->pipe_handle,
			       USB_FLAGS_SLEEP,NULL,NULL);
		*/
	}
	

#ifdef NOTYET
	printf("DEBUG: will call start_polling with softstate=%p\n",softstate);
	stat=usb_pipe_start_polling(softstate->pipe_handle,
				  NULL,
				  aipcam_poll_callback,
				  (void*)softstate);
#else
	stat=usb_pipe_start_polling(softstate->pipe_handle,
				  USB_FLAGS_SLEEP, NULL,NULL);
#endif

	if(stat != USB_SUCCESS)
	{
		cmn_err(CE_WARN,"start_polling of pipe FAILED");
		printusberr(stat);
		usb_pipe_close(&softstate->pipe_handle, USB_FLAGS_SLEEP, NULL,NULL);
		mutex_exit(&softstate->instance_mutex);
		return ENXIO;
	} else {
		printf("DEBUG: start_polling succeeded?");
	}

	if(usb_pipe_get_state(softstate->pipe_handle,
			      &pipestate,
			      USB_FLAGS_SLEEP) != USB_SUCCESS)
	{
		cmn_err(CE_WARN,"ERROR: could not get usb_pipe state!");
	}
	printf("DEBUG: pipe-state after start_polling is %d\n",pipestate);

	mutex_exit(&softstate->instance_mutex);


   	return 0;
}


static int aipcam_close(dev_t dev, int openflags, int otyp, cred_t *credp){
	int instance=DEVT_TO_INST(dev);
	struct aipcam_softstate *softstate;


	softstate=ddi_get_soft_state(aipcam_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"aipcam_close failed: cannot get softstate?");
		return DDI_FAILURE;
	}

   	aipcam_debugprint("aipcam: close called\n");

	/* Since we only allow a single process to open() at one time,
	 * and since the OS should not allow multiple calls to open()
	 * to actually get through to us... This routine is pseudo-thread-safe
	 * without using any mutexes.
	 */


	if(softstate->openflags==0){
		cmn_err(CE_WARN,"ERROR: aipcam_close: pipe not open");
		return ENXIO;
		
	}

	aipcam_writereg(softstate,OV511_REG_RESET, OV511_RESET_STOPCAM);

	/* Should not return until polling has stopped. */
	/* MUST NOT HOLD MUTEX HERE, otherwise we deadlock,
	 * due to isoc_callback using mutex.
	 */
	usb_pipe_stop_polling(softstate->pipe_handle,
			      USB_FLAGS_SLEEP,NULL,NULL);
	if(usb_pipe_close(&softstate->pipe_handle,
			  USB_FLAGS_SLEEP,NULL,NULL ) != USB_SUCCESS)
	{
		aipcam_debugprint("ERROR: usb_pipe_close failed?");
		return ENXIO;
	}
	softstate->pipe_handle=NULL;

	/* Okay, a little redundant since we do this in open(), but
	 * thats okay. Clear out the read queue.
	 */
	readqueue_reset(softstate);

	/* This is almost like a mutex */
	softstate->openflags=0;

	printf("DEBUG: close() returning successfully\n");
   	return (0);
}

static int aipcam_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
		    cred_t *credp, int *rvalp){
/*
	int instance=DEVT_TO_INST(dev);
	struct aipcam_softstate *softstate;
*/

	aipcam_debugprint("aipcam: ioctl called\n");
   	return (ENXIO);
}

static int aipcam_detach(dev_info_t *dip, ddi_detach_cmd_t cmd){
   	int instance;
	struct aipcam_softstate *softstate;
   	instance=ddi_get_instance(dip);

	aipcam_debugprint("aipcam: detach for instance %d\n", instance);

	softstate=ddi_get_soft_state(aipcam_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"aipcam_detach failed: cannot get softstate?");
		return DDI_FAILURE;
	}

	aipcam_cleanup(softstate);

	ddi_remove_minor_node(dip, DEVNODE_NAME);

	return (DDI_SUCCESS);
		     
}
